// BSP.cpp: implementation of the BSP class.
//
//////////////////////////////////////////////////////////////////////

#include "windows.h"
#include "BSPtree.h"


#define BSP_SEUIL 5


BSPTree::BSPTree(void)
{
	m_pRoot = NULL;
	m_pNodesByID = NULL;
	m_pOrphansByID = NULL;
}

BSPTree::~BSPTree(void)
{
	if (m_pNodesByID != NULL)
	{
		delete(m_pNodesByID);
		m_pNodesByID = NULL;
	}
	if (m_pOrphansByID != NULL)
	{
		delete(m_pOrphansByID);
		m_pOrphansByID = NULL;
	}
}

void BSPTree::AddNode(BSPNode* pNode)
{
	if (m_pNodesByID == NULL)
	{
		m_pNodesByID = (BSPNode**)calloc(m_nTotalNodes, sizeof(BSPNode*));
	}

	if ((m_pRoot == NULL) && (pNode->isRoot()))
	{
		m_pRoot = pNode;
		//m_pNodesByID->appendElement(pNode, pNode->m_nIndex);
		m_pNodesByID[pNode->m_nIndex] = pNode;
	}
	else
	{
		// find parent
		BSPNode* pParent = (BSPNode*)m_pNodesByID[pNode->m_nParentIndex];
		if (pParent != NULL)
		{
			if (pParent->m_nFrontIndex == pNode->m_nIndex)
			{
				pParent->SetFrontNode(pNode);
				m_pNodesByID[pNode->m_nIndex] = pNode;
			}
			else if (pParent->m_nBackIndex == pNode->m_nIndex)
			{
				pParent->SetBackNode(pNode);
				m_pNodesByID[pNode->m_nIndex] = pNode;
			}
			else if (pParent->m_nOverlapIndex == pNode->m_nIndex)
			{
				pParent->SetOverlapNode(pNode);
				m_pNodesByID[pNode->m_nIndex] = pNode;
			}
		}
		else
		{
			m_pOrphansByID[pNode->m_nIndex] = pNode;
		}
	}
}

int BSPTree::SelectViewpoint(int maxdepth, float x, float y, float z, ZArray<sint32>* pObjID)
{
	int nObjID = Parse(0, maxdepth, m_pRoot, x, y, z, pObjID);
	return pObjID->GetLength();
}

int BSPTree::SelectObjects(float x, float y, float z, ZArray<sint32>* pObjID)
{
	int nObjID = ParseObjects(0, m_pRoot, x, y, z, pObjID);
	return pObjID->GetLength();
}

int BSPTree::SelectOrientation(float x, float y, float z, float vX, float vY, float vZ, ZArray<sint32>* pObjID)
{
//	int nObjID = ParseOrientation(0, maxdepth, m_pRoot, x, y, z, vX, vY, vZ, pObjID);
	return 0;
}

int BSPTree::SelectFrustrum(int maxdepth, float* x, float* y, float* z, ZArray<sint32>* pObjID)
{
//	int nObjID = Parse(0, maxdepth, m_pRoot, x, y, z, pObjID);
	int nObjID = ParseFrustrum(m_pRoot, x, y, z, pObjID);
	return pObjID->GetLength();
}


int BSPTree::Parse(int depth, int maxdepth, BSPNode* pNode, float x, float y, float z, ZArray<sint32>* pObjID)
{
	if (pNode == NULL) return 0;

	if (depth == maxdepth)
	{
		return FillAllChildren(pNode, pObjID);
	}

	// test position of viewpoint, to plane
	float S = pNode->m_a*x + pNode->m_b*y + pNode->m_c*z + pNode->m_d;

	if (S <= 0)
	{
		// viewpoint on front of plane
		if (pNode->m_nFrontIndex > 0)
		{
			// check if FrontNode is there
			if (pNode->m_pFrontNode != NULL)
			{
				return Parse(depth+1, maxdepth, pNode->m_pFrontNode, x, y, z, pObjID);
			}
			else
			{
				BSPNode* pOrphan = (BSPNode*)m_pOrphansByID[pNode->m_nFrontIndex];
				// orphan found, then add to tree
				if (pOrphan != NULL)
				{
					pNode->SetFrontNode(pOrphan);
					m_pOrphansByID[pNode->m_nFrontIndex] = NULL;
					m_pNodesByID[pOrphan->m_nIndex] = pOrphan;
					return Parse(depth+1, maxdepth, pOrphan, x, y, z, pObjID);
				}
			}
		}
		else
		{
			if (pNode->m_nFrontObjCount > 0)
			{
				for (unsigned int iObj=0; iObj<pNode->m_nFrontObjCount; iObj++)
				{
					int nObjID = pNode->m_lFrontObjID[iObj];
					pObjID->InsertAt(pObjID->GetLength(), (nObjID));
				}
			}
		}
	}
	else if (S > 0)
	{
		// viewpoint on back of plane
		if (pNode->m_nBackIndex > 0)
		{
			// check if BackNode is there
			if (pNode->m_pBackNode != NULL)
			{
				return Parse(depth+1, maxdepth, pNode->m_pBackNode, x, y, z, pObjID);
			}
			else
			{
				BSPNode* pOrphan = (BSPNode*)m_pOrphansByID[pNode->m_nBackIndex];
				// orphan found, then add to tree
				if (pOrphan != NULL)
				{
					pNode->SetBackNode(pOrphan);
					m_pOrphansByID[pNode->m_nBackIndex] = NULL; 
					m_pNodesByID[pOrphan->m_nIndex] = pOrphan;
					return Parse(depth+1, maxdepth, pOrphan, x, y, z, pObjID);
				}
			}
		}
		else
		{
			if (pNode->m_nBackObjCount > 0)
			{
				for (int iObj=0; iObj<pNode->m_nBackObjCount; iObj++)
				{
					int nObjID = pNode->m_lBackObjID[iObj];
					pObjID->InsertAt(pObjID->GetLength(), (nObjID));
				}
			}
		}
	}
/*	else
	{
		// case on plane????
	}
	*/
	return pObjID->GetLength();
}

int BSPTree::ParseObjects(int depth, BSPNode* pNode, float x, float y, float z, ZArray<sint32>* pObjID)
{
	if (pNode == NULL) return 0;

	// test position of viewpoint, to plane
	float S = pNode->m_a*x + pNode->m_b*y + pNode->m_c*z + pNode->m_d;

	if (S < -BSP_SEUIL)
	{
		ParseFrontObjects(depth, pNode, x, y, z, pObjID);
		ParseOverlapObjects(depth, pNode, x, y, z, pObjID);
		ParseBackObjects(depth, pNode, x, y, z, pObjID);
	}

	// parse overlap
	else if (S > BSP_SEUIL)
	{
		ParseBackObjects(depth, pNode, x, y, z, pObjID);
		ParseOverlapObjects(depth, pNode, x, y, z, pObjID);
		ParseFrontObjects(depth, pNode, x, y, z, pObjID);
	}
	else
	{
		if (S > 0)
		{
			ParseOverlapObjects(depth, pNode, x, y, z, pObjID);
			ParseBackObjects(depth, pNode, x, y, z, pObjID);
			ParseFrontObjects(depth, pNode, x, y, z, pObjID);
		}
		else
		{
			ParseOverlapObjects(depth, pNode, x, y, z, pObjID);
			ParseFrontObjects(depth, pNode, x, y, z, pObjID);
			ParseBackObjects(depth, pNode, x, y, z, pObjID);
		}
	}

	return pObjID->GetLength();
}

int BSPTree::ParseFrontObjects(int depth, BSPNode* pNode, float x, float y, float z, ZArray<sint32>* pObjID)
{
	// viewpoint on front of plane
	if (pNode->m_nFrontIndex > 0)
	{
		// check if FrontNode is there
		if (pNode->m_pFrontNode != NULL)
		{
			return ParseObjects(depth+1, pNode->m_pFrontNode, x, y, z, pObjID);
		}
		else
		{
			BSPNode* pOrphan = (BSPNode*)m_pOrphansByID[pNode->m_nFrontIndex];
			// orphan found, then add to tree
			if (pOrphan != NULL)
			{
				pNode->SetFrontNode(pOrphan);
				m_pOrphansByID[pNode->m_nFrontIndex] = NULL;
				m_pNodesByID[pOrphan->m_nIndex] = pOrphan;
				return ParseObjects(depth+1, pOrphan, x, y, z, pObjID);
			}
		}
	}
	else
	{
		if (pNode->m_nFrontObjCount > 0)
		{
			for (int iObj=0; iObj<pNode->m_nFrontObjCount; iObj++)
			{
				int nObjID = pNode->m_lFrontObjID[iObj];
				pObjID->InsertAt(pObjID->GetLength(), (nObjID));
			}
		}
	}

	return pObjID->GetLength();
}


int BSPTree::ParseOverlapObjects(int depth, BSPNode* pNode, float x, float y, float z, ZArray<sint32>* pObjID)
{
	// viewpoint on back of plane
	if (pNode->m_nOverlapIndex > 0)
	{
		// check if OverlapNode is there
		if (pNode->m_pOverlapNode != NULL)
		{
			return ParseObjects(depth+1, pNode->m_pOverlapNode, x, y, z, pObjID);
		}
		else
		{
			BSPNode* pOrphan = (BSPNode*)m_pOrphansByID[pNode->m_nOverlapIndex];
			// orphan found, then add to tree
			if (pOrphan != NULL)
			{
				pNode->SetOverlapNode(pOrphan);
				m_pOrphansByID[pNode->m_nOverlapIndex] = NULL;
				m_pNodesByID[pOrphan->m_nIndex] = pOrphan;
				return ParseObjects(depth+1, pOrphan, x, y, z, pObjID);
			}
		}
	}
	else
	{
		if (pNode->m_nOverlapObjCount > 0)
		{
			for (int iObj=0; iObj<pNode->m_nOverlapObjCount; iObj++)
			{
				int nObjID = pNode->m_lOverlapObjID[iObj];
				pObjID->InsertAt(pObjID->GetLength(), (nObjID));
			}
		}
	}

	return pObjID->GetLength();
}


int BSPTree::ParseBackObjects(int depth, BSPNode* pNode, float x, float y, float z, ZArray<sint32>* pObjID)
{
	// viewpoint on back of plane
	if (pNode->m_nBackIndex > 0)
	{
		// check if BackNode is there
		if (pNode->m_pBackNode != NULL)
		{
			return ParseObjects(depth+1, pNode->m_pBackNode, x, y, z, pObjID);
		}
		else
		{
			BSPNode* pOrphan = (BSPNode*)m_pOrphansByID[pNode->m_nBackIndex];
			// orphan found, then add to tree
			if (pOrphan != NULL)
			{
				pNode->SetBackNode(pOrphan);
				m_pOrphansByID[pNode->m_nBackIndex] = NULL;
				m_pNodesByID[pOrphan->m_nIndex] = pOrphan;
				return ParseObjects(depth+1, pOrphan, x, y, z, pObjID);
			}
		}
	}
	else
	{
		if (pNode->m_nBackObjCount > 0)
		{
			for (int iObj=0; iObj<pNode->m_nBackObjCount; iObj++)
			{
				int nObjID = pNode->m_lBackObjID[iObj];
				pObjID->InsertAt(pObjID->GetLength(), (nObjID));
			}
		}
	}

	return pObjID->GetLength();
}


// used with frustrum
int BSPTree::Parse(int depth, int maxdepth, BSPNode* pNode, float* x, float* y, float* z, ZArray<sint32>* pObjID)
{
	if (pNode == NULL) return 0;

	if (depth == maxdepth)
	{
		return FillAllChildren(pNode, pObjID);
	}

	// test frustrum position, to plane
	bool bFront=false;
	bool bBack= false;
	for (int ipoint=0; ipoint<8; ipoint++)
	{
		float S = pNode->m_a*x[ipoint] + pNode->m_b*y[ipoint] + pNode->m_c*z[ipoint] + pNode->m_d;
		if (S>0) bBack=true;
		if (S<=0) bFront=true;
	}

	if (bFront)
	{
		// frustrum on front of plane
		if (pNode->m_nFrontIndex > 0)
		{
			// check if FrontNode is there
			if (pNode->m_pFrontNode != NULL)
			{
				return Parse(depth+1, maxdepth, pNode->m_pFrontNode, x, y, z, pObjID);
			}
			else
			{
				BSPNode* pOrphan = (BSPNode*)m_pOrphansByID[pNode->m_nFrontIndex];
				// orphan found, then add to tree
				if (pOrphan != NULL)
				{
					pNode->SetFrontNode(pOrphan);
					m_pOrphansByID[pNode->m_nFrontIndex] = NULL;
					m_pNodesByID[pOrphan->m_nIndex] = pOrphan;
					return Parse(depth+1, maxdepth, pOrphan, x, y, z, pObjID);
				}
			}
		}
		else
		{
			if (pNode->m_nFrontObjCount > 0)
			{
				for (int iObj=0; iObj<pNode->m_nFrontObjCount; iObj++)
				{
					int nObjID = pNode->m_lFrontObjID[iObj];
					pObjID->InsertAt(pObjID->GetLength(), (nObjID));
				}
			}
		}
	}
	if (bBack)
	{
		// viewpoint on back of plane
		if (pNode->m_nBackIndex > 0)
		{
			// check if BackNode is there
			if (pNode->m_pBackNode != NULL)
			{
				return Parse(depth+1, maxdepth, pNode->m_pBackNode, x, y, z, pObjID);
			}
			else
			{
				BSPNode* pOrphan = (BSPNode*)m_pOrphansByID[pNode->m_nBackIndex];
				// orphan found, then add to tree
				if (pOrphan != NULL)
				{
					pNode->SetBackNode(pOrphan);
					m_pOrphansByID[pNode->m_nBackIndex] = NULL;
					m_pNodesByID[pOrphan->m_nIndex] = pOrphan;
					return Parse(depth+1, maxdepth, pOrphan, x, y, z, pObjID);
				}
			}
		}
		else
		{
			if (pNode->m_nBackObjCount > 0)
			{
				for (int iObj=0; iObj<pNode->m_nBackObjCount; iObj++)
				{
					int nObjID = pNode->m_lBackObjID[iObj];
					pObjID->InsertAt(pObjID->GetLength(), (nObjID));
				}
			}
		}
	}
/*	else
	{
		// case on plane????
	}
	*/
	return pObjID->GetLength();
}

int BSPTree::ParseFrustrum(BSPNode* pNode, float* x, float* y, float* z, ZArray<sint32>* pObjID)
{
	if (pNode == NULL) return 0;


	// test frustrum position, to plane
	bool bFront=false;
	bool bBack= false;
	for (int ipoint=0; ipoint<8; ipoint++)
	{
		float S = pNode->m_a*x[ipoint] + pNode->m_b*y[ipoint] + pNode->m_c*z[ipoint] + pNode->m_d;
		if (S>0) bBack=true;
		if (S<=0) bFront=true;
	}

	if (bFront)
	{
		// frustrum on front of plane
		if (pNode->m_nFrontIndex > 0)
		{
			// check if FrontNode is there
			if (pNode->m_pFrontNode != NULL)
			{
				return ParseFrustrum(pNode->m_pFrontNode, x, y, z, pObjID);
			}
			else
			{
				BSPNode* pOrphan = (BSPNode*)m_pOrphansByID[pNode->m_nFrontIndex];
				// orphan found, then add to tree
				if (pOrphan != NULL)
				{
					pNode->SetFrontNode(pOrphan);
					m_pOrphansByID[pNode->m_nFrontIndex] = NULL;
					m_pNodesByID[pOrphan->m_nIndex] = pOrphan;
					return ParseFrustrum(pOrphan, x, y, z, pObjID);
				}
			}
		}
		else
		{
			if (pNode->m_nFrontObjCount > 0)
			{
				for (int iObj=0; iObj<pNode->m_nFrontObjCount; iObj++)
				{
					int nObjID = pNode->m_lFrontObjID[iObj];
					pObjID->InsertAt(pObjID->GetLength(), (nObjID));
				}
			}
		}
	}
	if (bBack)
	{
		// viewpoint on back of plane
		if (pNode->m_nBackIndex > 0)
		{
			// check if BackNode is there
			if (pNode->m_pBackNode != NULL)
			{
				return ParseFrustrum(pNode->m_pBackNode, x, y, z, pObjID);
			}
			else
			{
				BSPNode* pOrphan = (BSPNode*)m_pOrphansByID[pNode->m_nBackIndex];
				// orphan found, then add to tree
				if (pOrphan != NULL)
				{
					pNode->SetBackNode(pOrphan);
					m_pOrphansByID[pNode->m_nBackIndex] = NULL;
					m_pNodesByID[pOrphan->m_nIndex] = pOrphan;
					return ParseFrustrum(pOrphan, x, y, z, pObjID);
				}
			}
		}
		else
		{
			if (pNode->m_nBackObjCount > 0)
			{
				for (int iObj=0; iObj<pNode->m_nBackObjCount; iObj++)
				{
					int nObjID = pNode->m_lBackObjID[iObj];
					pObjID->InsertAt(pObjID->GetLength(), (nObjID));
				}
			}
		}
	}
/*	else
	{
		// case on plane????
	}
	*/
	return pObjID->GetLength();
}


int BSPTree::FillAllChildren(BSPNode* pNode, ZArray<sint32>* pObjID)
{
	if (pNode->m_nBackObjCount > 0)
	{
		// fill back objects list
		for (int iObj=0; iObj<pNode->m_nBackObjCount; iObj++)
		{
			int nID = pNode->m_lBackObjID[iObj];
			pObjID->InsertAt(pObjID->GetLength() ,nID);
		}
	}

	if (pNode->m_nFrontObjCount > 0)
	{
		// fill front objects list
		for (int iObj=0; iObj<pNode->m_nFrontObjCount; iObj++)
		{
			int nID = pNode->m_lFrontObjID[iObj];
			pObjID->InsertAt(pObjID->GetLength() ,nID);
		}
	}

	if (pNode->m_nBackIndex > 0)
	{
		// check if BackNode is there
		if (pNode->m_pBackNode != NULL)
		{
			FillAllChildren(pNode->m_pBackNode, pObjID);
		}
		else
		{
			BSPNode* pOrphan = (BSPNode*)m_pOrphansByID[pNode->m_nBackIndex];
			// orphan found, then add to tree
			if (pOrphan != NULL)
			{
				pNode->SetBackNode(pOrphan);
				m_pOrphansByID[pNode->m_nBackIndex] = NULL;
				m_pNodesByID[pOrphan->m_nIndex] = NULL;
				FillAllChildren(pNode->m_pBackNode, pObjID);
			}
		}
	}

	if (pNode->m_nFrontIndex > 0)
	{
		// check if FrontNode is there
		if (pNode->m_pFrontNode != NULL)
		{
			FillAllChildren(pNode->m_pFrontNode, pObjID);
		}
		else
		{
			BSPNode* pOrphan = (BSPNode*)m_pOrphansByID[pNode->m_nFrontIndex];
			// orphan found, then add to tree
			if (pOrphan != NULL)
			{
				pNode->SetFrontNode(pOrphan);
				m_pOrphansByID[pNode->m_nFrontIndex] = NULL;
				m_pNodesByID[pOrphan->m_nIndex] = pOrphan;
				FillAllChildren(pNode->m_pFrontNode, pObjID);
			}
		}
	}

	if (pNode->m_pFrontNode != NULL)
	{
		FillAllChildren(pNode->m_pFrontNode, pObjID);
	}

	return pObjID->GetLength();
}
